/*
								+---------------------------------+
								|                                 |
								|       ***   Matice   ***        |
								|                                 |
								|  Copyright   -tHE SWINe- 2002  |
								|                                 |
								|           Matrix.cpp            |
								|                                 |
								+---------------------------------+
*/

#include <windows.h>
#include <malloc.h>
#include <math.h>
#include "Matrix.h"

//---------------------------------------------------
// vytvori jednotkovou matici

void Matrix_Init(Matrix *m)
{
    int i, j;

    for(i = 0; i < 4; i ++)
        for(j = 0; j < 4; j ++)
            (*m)[i][j] = (float)(i == j);
}


//---------------------------------------------------
// nasobeni dvou matic

void Matrix_Multiply(Matrix *a, Matrix *b, Matrix *r)
{
    Matrix temp;
    int i, j;

    for(i = 0; i < 4; i ++) {
        for(j = 0; j < 4; j ++)
            temp[i][j] = (*a)[0][j] * (*b)[i][0] + (*a)[1][j] * (*b)[i][1] +
                (*a)[2][j] * (*b)[i][2] + (*a)[3][j] * (*b)[i][3];
    }
    memcpy(r, temp, sizeof(Matrix));
}


//---------------------------------------------------
// rotace matice po ose x

void Matrix_Rotate_X(Matrix *m, float angle)
{
    Matrix temp;

    Matrix_Init(&temp);

    temp[2][1] = (float)-sin(angle);
    temp[2][2] = (float)cos(angle);
    temp[1][1] = (float)cos(angle);
    temp[1][2] = (float)sin(angle);

    Matrix_Multiply(m, &temp, m);
}


//---------------------------------------------------
// rotace matice po ose y

void Matrix_Rotate_Y(Matrix *m, float angle)
{
    Matrix temp;

    Matrix_Init(&temp);

    temp[0][0] = (float)cos(angle);
    temp[0][2] = (float)-sin(angle);
    temp[2][2] = (float)cos(angle);
    temp[2][0] = (float)sin(angle);

    Matrix_Multiply(m, &temp, m);
}


//---------------------------------------------------
// rotace matice po ose z

void Matrix_Rotate_Z(Matrix *m, float angle)
{
    Matrix temp;

    Matrix_Init(&temp);

    temp[0][0] = (float)cos(angle);
    temp[0][1] = (float)sin(angle);
    temp[1][1] = (float)cos(angle);
    temp[1][0] = (float)-sin(angle);

    Matrix_Multiply(m, &temp, m);
}


//---------------------------------------------------
// posunuti matice

void Matrix_Translate(Matrix * m, float dx, float dy, float dz)
{
    Matrix temp;

    Matrix_Init(&temp);

    temp[3][0] = dx;
    temp[3][1] = dy;
    temp[3][2] = dz;

    Matrix_Multiply(m, &temp, m);
}


//---------------------------------------------------
// spocita inverzni matici

void Matrix_Inverse(Matrix *src, Matrix *inv)
{
    float f_det;
    int i, j;

    f_det = (*src)[0][0] * (*src)[1][1] * (*src)[2][2] +
            (*src)[1][0] * (*src)[2][1] * (*src)[0][2] +
            (*src)[0][1] * (*src)[1][2] * (*src)[2][0] -
            (*src)[0][2] * (*src)[1][1] * (*src)[2][0] -
            (*src)[0][1] * (*src)[1][0] * (*src)[2][2] -
            (*src)[0][0] * (*src)[1][2] * (*src)[2][1];
    // spocita determinant podle sarrusova pravidla
    // u matice se spodkem "0 0 0 1" je determinant
    // stejny jako posledni doplnek

    (*inv)[0][0] = (*src)[1][1] * (*src)[2][2] /** (*src)[3][3] +*/ // 33 = 1
                   /*(*src)[2][1] * (*src)[3][2] * (*src)[1][3] +*/ // 13 = 0
                   /*(*src)[1][2] * (*src)[2][3] * (*src)[3][1] -*/ // 23 = 0
                   /*(*src)[3][1] * (*src)[2][2] * (*src)[1][3] -*/ // 13 = 0
                   /*(*src)[1][1] * (*src)[2][3] * (*src)[3][2] -*/ // 23 = 0
                   -(*src)[1][2] * (*src)[2][1] /** (*src)[3][3]*/; // 33 = 1
    (*inv)[1][0] =-(*src)[1][0] * (*src)[2][2] /** (*src)[3][3] +*/ // 33 = 1
                   /*(*src)[2][0] * (*src)[3][2] * (*src)[1][3] +*/ // 13 = 0
                   /*(*src)[1][2] * (*src)[2][3] * (*src)[3][0] -*/ // 23 = 0
                   /*(*src)[1][3] * (*src)[2][2] * (*src)[3][0] -*/ // 13 = 0
                   /*(*src)[2][3] * (*src)[3][2] * (*src)[1][0] -*/ // 23 = 0
                   +(*src)[1][2] * (*src)[2][0] /** (*src)[3][3]*/; // 33 = 1
    (*inv)[2][0] = (*src)[1][0] * (*src)[2][1] /** (*src)[3][3] +*/ // 33 = 1
                   /*(*src)[1][3] * (*src)[2][0] * (*src)[3][1] +*/ // 13 = 0
                   /*(*src)[2][3] * (*src)[1][1] * (*src)[3][1] -*/ // 23 = 0
                   /*(*src)[1][3] * (*src)[2][1] * (*src)[3][0] -*/ // 13 = 0
                   /*(*src)[2][3] * (*src)[3][2] * (*src)[1][0] -*/ // 23 = 0
                   -(*src)[1][1] * (*src)[2][0] /** (*src)[3][3]*/; // 33 = 1
    (*inv)[3][0] =-(*src)[1][0] * (*src)[2][1] * (*src)[3][2] -
                   (*src)[1][1] * (*src)[2][2] * (*src)[3][0] -
                   (*src)[2][0] * (*src)[3][1] * (*src)[1][2] +
                   (*src)[1][2] * (*src)[2][1] * (*src)[3][0] +
                   (*src)[1][1] * (*src)[2][0] * (*src)[3][2] +
                   (*src)[2][2] * (*src)[1][0] * (*src)[3][1];
    // takhle moc se to vecko zkrt, dl u to nebudu rozepisovat ...

    (*inv)[0][1] =-(*src)[0][1] * (*src)[2][2] + (*src)[0][2] * (*src)[2][1];
    (*inv)[1][1] = (*src)[0][0] * (*src)[2][2] - (*src)[0][2] * (*src)[2][0];
    (*inv)[2][1] =-(*src)[0][0] * (*src)[2][1] + (*src)[0][1] * (*src)[2][0];
    (*inv)[3][1] = (*src)[0][0] * (*src)[2][1] * (*src)[3][2] +
                   (*src)[0][1] * (*src)[2][2] * (*src)[3][0] +
                   (*src)[2][0] * (*src)[3][1] * (*src)[0][2] -
                   (*src)[0][2] * (*src)[2][1] * (*src)[3][0] -
                   (*src)[0][1] * (*src)[2][0] * (*src)[3][2] -
                   (*src)[2][2] * (*src)[0][0] * (*src)[3][1];
    (*inv)[0][2] = (*src)[0][1] * (*src)[1][2] - (*src)[0][2] * (*src)[1][1];
    (*inv)[1][2] =-(*src)[0][0] * (*src)[1][2] + (*src)[0][2] * (*src)[1][0];
    (*inv)[2][2] = (*src)[0][0] * (*src)[1][1] - (*src)[0][1] * (*src)[1][0];
    (*inv)[3][2] =-(*src)[0][0] * (*src)[1][1] * (*src)[3][2] -
                   (*src)[0][1] * (*src)[1][2] * (*src)[3][0] -
                   (*src)[1][0] * (*src)[3][1] * (*src)[0][2] +
                   (*src)[0][2] * (*src)[1][1] * (*src)[3][0] +
                   (*src)[0][1] * (*src)[1][0] * (*src)[3][2] +
                   (*src)[1][2] * (*src)[0][0] * (*src)[3][1];

    for(j = 0; j < 4; j ++)
        for(i = 0; i < 3; i ++)
            (*inv)[j][i] /= f_det;
    // vecko vydl determinantem (krom poslednho sloupce)

    (*inv)[0][3] = 0;
    (*inv)[1][3] = 0;
    (*inv)[2][3] = 0;
    (*inv)[3][3] = 1;
    // pi matici se spodnm dkem "0 0 0 1" vyjde vdycky posledn sloupec takhle.
}